# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import opteryx.virtual_datasets.astronaut_data as astronauts
import opteryx.virtual_datasets.derived_data as derived
import opteryx.virtual_datasets.no_table_data as no_table
import opteryx.virtual_datasets.planet_data as planets
import opteryx.virtual_datasets.satellite_data as satellites
import opteryx.virtual_datasets.variables_data as variables
